(*****************************************************************************
 *  Author:
 *    Daniel N. Gtzmann <dngoetzmann@googlemail.com>
 *
 *  Copyright:
 *     Daniel N. Gtzmann, 2009
 *
 *  Last modified:
 *    $Date: 2009-01-30 16:26:36 +0100 (Fri, 30 Jan 2009) $
 *    $Author: goetzmann $
 *    $Revision: 373 $
 *
 *  This file is part of Spartacus,
 *  the tableau prover for hybrid logic
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *****************************************************************************)


val _ =
	let
		val _ = Solver.startTimer ()
		
		exception Help
		exception InvalidArgument
		
		val helpMsg =
			"Usage: spartacus [OPTIONS]...\n" ^
			"\nInput:\n" ^
			"--formula=FORMULA        use the given FORMULA as input\n" ^
			"--file=FILE              read formula from FILE\n" ^
			"--dfgFile=FILE           read formula in dfg-ascii format from FILE\n" ^
			"--dimacsFile=FILE        read formula in DIMACS format from FILE\n" ^
			"--dl98File=FILE          read formula in DL98 format from FILE\n" ^
			"--intohyloFile=FILE      read formula in InToHyLo format from FILE\n" ^
			"--ksatcFile=FILE         read formula in KSatC format from FILE\n" ^
			"--ksatlFile=FILE         read formula in KSatLisp format from FILE\n" ^
			"--lwbFile=FILE           read formula in LWB format from FILE\n" ^
			"--tancsFile=FILE         read formula in TANCS format from FILE\n" ^
			"--index=n                run the prover on the nth formula (e.g. when using an LWB file as input)\n" ^
			"--negate                 run the prover on the negation of the given formula\n" ^
			"--reflexive              handle relations as reflexive\n" ^
			"--transitive             handle relations as transitive\n" ^
			"--serial                 handle relations as serial\n" ^
			"\nPredefined configuration profile:\n" ^
			"--profile=FILE           use options as specified in FILE\n" ^
			"\nOrder:\n" ^
			"--exp-ord=S              order in which rules are applied. S is a string containing\n" ^
			"                             n (nominals), N (negated nominals), < (diamond), [ (box),\n" ^
			"                             A (universal), E (existential), @ (satisfaction op.), | (disjunction).\n" ^
			"                             If x occurs before y in S, the rule for x will be applied\n" ^
			"                             before the rule for y.\n" ^
			"--disj-ord=S             order in which disjuncts are processed. S is a string containing\n" ^
			"                             p (propositions), n (nominals), N (negated nominals), < (diamond), [ (box),\n" ^
			"                             A (universal), E (existential), @ (satisfaction op.), & (conjunction).\n" ^
			"                             If x occurs before y in S, disjuncts of type x will be considered\n" ^
			"                             before terms of type y.\n" ^
			"--dia-old                consider diamonds on old nodes first\n" ^
			"--dia-new                consider diamonds on new nodes first (default)\n" ^
			"--dia-lifo               consider (globally) new diamonds first first\n" ^
			"--dia-fifo               consider (globally) old diamonds first\n" ^
			"--dia-dep                consider diamonds whose dependency sets contain least recent branching points first\n" ^
			"--disj-old               consider disjunctions on old nodes first (default)\n" ^
			"--disj-new               consider disjunctions on new nodes first\n" ^
			"--disj-lifo              consider (globally) new disjunctions first\n" ^
			"--disj-fifo              consider (globally) old disjunctions first\n" ^
			"--disj-dep               consider disjunctions whose dependency sets contain least recent branching points first\n" ^
			"\nOptimizations:\n" ^
			"--backjumping=[off|on]    turn backjumping off/on (default: on)\n" ^
			"--blocking=[off|on|eager] turn pattern-based blocking off/on/on including new boxes (default: eager)\n" ^
			"--caching=[off|on|eager]  turn caching off/on/on with additional cache queries when propagating (default: off)\n" ^
			"--bcp=[off|on|eager]      turn boolean constraint propagation off/on considering only next disjunction/on searching for disjunctions (default: eager)\n" ^
			"--ecd=[off|on]            turn early conflict detection off/on (default: on)\n" ^
			"--db=[off|ngl|on]         turn disjoint branching off/on using no good lists only/on (default: on)\n" ^
			"--lazy=[off|box|prop|on]  turn lazy branching off/boxes only/propositional literals only/on (default: on)" ^
			"\nData structures:\n" ^
			"--blockingds=[list|tree|matrix] use list-based/tree-based/bitmatrix-based data structure for pattern-based blocking (default: tree)\n" ^
			"--cachingds=[tree|matrix]       use tree-based/bitmatrix-based data structure for caching (default: tree)\n" ^
			"--cachingds=matrix:N            use size-bounded bitmatrix-based data structure for caching with size bound N\n" ^
			"\nOutput:\n" ^
			"--dotFile=FILE           save a model in dot format to FILE if the formula is satisfiable\n" ^
			"--showModel              print a model if the formula is satisfiable\n" ^
			"--dmv                    print a detailed view for each node in the model\n" ^
			"--snc                    show negative constraints in output\n" ^
			"--debug                  print debug output\n" ^
			"\n" ^
			"--timeout=n              run for at most n seconds\n" ^
			"--csv                    output as comma-separated list\n"
		
		
		val csvHeader = "file name, index, --negate?, exp-ord, disj-ord, --dia-X, --disj-X, backjumping, blocking, caching, bcp, ecd, db, lazy, blocking ds, caching ds, timeout, parse time, nnf translation time, indexing time, initialization time, decision time, branches, total time, result\n"
		
		
		datatype inputFormat = ARG | NAT | LWB | KSATC | ALC | INTOHYLO | DIMACS | DFG | DL98 | TANCS
		
		val format = ref ARG
		val formula = ref ""
		val fileName = ref ""
		val index = ref 1
		val dotFileName = ref NONE
		
		val negate = ref false
		
		val showModel = ref false
		
		val parsetree = ref (Parsetree.DISJ [])
		
		val args = CommandLine.arguments ()
		
		fun readargs "--help" = (print helpMsg; raise Help)
		  | readargs "--csv" = Settings.csv := true
		  | readargs "--csvheader" = (print csvHeader; raise Help)
		  | readargs "--caching=on" = (Settings.cachingEnabled := true; Settings.checkCacheOften := false)
		  | readargs "--caching=off" = Settings.cachingEnabled := false
		  | readargs "--caching=eager" = (Settings.cachingEnabled := true; Settings.checkCacheOften := true)
		  | readargs "--blocking=on" = (Settings.pbBlockingEnabled := true; Settings.fullBlocking := false)
		  | readargs "--blocking=off" = (Settings.pbBlockingEnabled := false; Settings.fullBlocking := false)
		  | readargs "--blocking=eager" = (Settings.pbBlockingEnabled := true; Settings.fullBlocking := true)
		  | readargs "--ecd=on" = Settings.ecdDisabled := false
		  | readargs "--ecd=off" = Settings.ecdDisabled := true
		  | readargs "--db=on" = (Settings.semanticBranchingEnabled := true; Settings.semanticBranchingNgl := false)
		  | readargs "--db=ngl" = (Settings.semanticBranchingEnabled := true; Settings.semanticBranchingNgl := true)
		  | readargs "--db=off" = (Settings.semanticBranchingEnabled := false; Settings.semanticBranchingNgl := false)
		  | readargs "--lazy=prop" = (Settings.lazyBranching := true; Settings.lazyWithBoxes := false)
		  | readargs "--lazy=off" = (Settings.lazyBranching := false; Settings.lazyWithBoxes := false)
		  | readargs "--lazy=on" = (Settings.lazyBranching := true; Settings.lazyWithBoxes := true)
		  | readargs "--lazy=box" = (Settings.lazyBranching := false; Settings.lazyWithBoxes := true)
		  | readargs "--debug" = Debug.outputEnabled := true
		  | readargs "--xdb" = Agenda.setOrder "<[nAE@|"
		  | readargs "--xbd" = Agenda.setOrder "[nAE@|<"
		  | readargs "--showModel" = showModel := true
		  | readargs "--dmv" = Settings.detailedModelView := true
		  | readargs "--snc" = Settings.showNegativeConstraints := true
		  | readargs "--negate" = negate := true
		  | readargs "--backjumping=on" = Settings.backjumpingDisabled := false
		  | readargs "--backjumping=off" = Settings.backjumpingDisabled := true
		  | readargs "--bcp=eager" = (Settings.bcpEnabled := true; Settings.bcpFullyEnabled := true)
		  | readargs "--bcp=on" = (Settings.bcpEnabled := true; Settings.bcpFullyEnabled := false)
		  | readargs "--bcp=off" = (Settings.bcpEnabled := false; Settings.bcpFullyEnabled := false)
		  | readargs "--dia-old" = Settings.diaExpOrder := Settings.OLDWORLD
		  | readargs "--dia-new" = Settings.diaExpOrder := Settings.NEWWORLD
		  | readargs "--dia-lifo" = Settings.diaExpOrder := Settings.LIFO
		  | readargs "--dia-fifo" = Settings.diaExpOrder := Settings.FIFO
		  | readargs "--dia-dep" = Settings.diaExpOrder := Settings.DEP
		  | readargs "--disj-old" = Settings.disjExpOrder := Settings.OLDWORLD
		  | readargs "--disj-new" = Settings.disjExpOrder := Settings.NEWWORLD
		  | readargs "--disj-lifo" = Settings.disjExpOrder := Settings.LIFO
		  | readargs "--disj-fifo" = Settings.disjExpOrder := Settings.FIFO
		  | readargs "--disj-dep" = Settings.disjExpOrder := Settings.DEP
		  | readargs "--blockingds=list" = Settings.pbbDatastructure := Settings.LISTS
		  | readargs "--blockingds=tree" = Settings.pbbDatastructure := Settings.UBTREE
		  | readargs "--blockingds=matrix" = Settings.pbbDatastructure := Settings.BITMATRIX
		  | readargs "--cachingds=tree" = Settings.cacheDatastructure := Settings.UBTREE
		  | readargs "--cachingds=matrix" = Settings.cacheDatastructure := Settings.BITMATRIX
		  | readargs "--reflexive" = RelationMgr.setAllReflexive ()
		  | readargs "--transitive" = RelationMgr.setAllTransitive ()
		  | readargs "--serial" = RelationMgr.setAllSerial ()
		  | readargs s =
				if String.isPrefix "--formula=" s
				then formula := (String.extract (s, 10, NONE))
				else if String.isPrefix "--file=" s
				then (fileName := (String.extract (s, 7, NONE)); format := NAT)
				else if String.isPrefix "--lwbFile=" s
				then (fileName := (String.extract (s, 10, NONE)); format := LWB)
				else if String.isPrefix "--ksatcFile=" s
				then (fileName := (String.extract (s, 12, NONE)); format := KSATC)
				else if String.isPrefix "--alcFile=" s
				then (fileName := (String.extract (s, 10, NONE)); format := ALC)
				else if String.isPrefix "--ksatlFile=" s
				then (fileName := (String.extract (s, 12, NONE)); format := ALC)
				else if String.isPrefix "--intohyloFile=" s
				then (fileName := (String.extract (s, 15, NONE)); format := INTOHYLO)
				else if String.isPrefix "--dimacsFile=" s
				then (fileName := (String.extract (s, 13, NONE)); format := DIMACS)
				else if String.isPrefix "--dfgFile=" s
				then (fileName := (String.extract (s, 10, NONE)); format := DFG)
				else if String.isPrefix "--tkbFile=" s
				then (fileName := (String.extract (s, 10, NONE)); format := DL98)
				else if String.isPrefix "--dl98File=" s
				then (fileName := (String.extract (s, 11, NONE)); format := DL98)
				else if String.isPrefix "--tancsFile=" s
				then (fileName := (String.extract (s, 12, NONE)); format := TANCS)
				else if String.isPrefix "--index=" s
				then index := (Option.getOpt (Int.fromString (String.extract (s, 8, NONE)), !index))
				else if String.isPrefix "--lwbIndex=" s
				then index := (Option.getOpt (Int.fromString (String.extract (s, 11, NONE)), !index))
				else if String.isPrefix "--timeout=" s
				then Settings.timeout := SOME (Time.fromReal (Option.getOpt (Real.fromString (String.extract (s, 10, NONE)), 0.0)))
				else if String.isPrefix "--disj-ord=" s
				then Term.Catstore.setOrder (String.extract (s, 11, NONE))
				else if String.isPrefix "--exp-ord=" s
				then Agenda.setOrder (String.extract (s, 10, NONE))
				else if String.isPrefix "--dotFile=" s
				then dotFileName := (SOME (String.extract (s, 10, NONE)))
				else if String.isPrefix "--cachingds=matrix:" s
				then Settings.cacheDatastructure := (Settings.SIZEBOUNDED ((fn n => if n > 0 then n else (print "Invalid argument for cache size bound\n"; raise InvalidArgument)) (Option.getOpt (Int.fromString (String.extract (s, 19, NONE)), !index))))
				else if String.isPrefix "--profile=" s
				then
					let
						val f = TextIO.openIn (String.extract (s, 10, NONE))
						
						fun ignoreComment s = if String.isPrefix "//" s then "" else s
						
						fun readall () =
							if TextIO.endOfStream f
							then ()
							else (
								  app
									readargs
									(String.tokens
										(fn c => c = #" " orelse c = #"\t" orelse c = #"\n")
										(ignoreComment (Option.valOf (TextIO.inputLine f)))
									)
								; readall ()
							)
					in
						readall ()
					end
				else (print ("Invalid argument: " ^ s ^ "\n"); raise InvalidArgument)
		
		fun printSettings () =
			print (
				  "backjumping: "
				^ (if !Settings.backjumpingDisabled then "off" else "on")
				^ "\npattern-based blocking: "
				^ (if !Settings.pbBlockingEnabled then (if !Settings.fullBlocking then "eager" else "on") ^ " (using " ^ (case !Settings.pbbDatastructure of Settings.LISTS => "list" | Settings.UBTREE => "tree" | Settings.BITMATRIX => "bitmatrix" | _ => Exn.unexpected "Prover.printSettings: unexpected datastructure for blocking") ^ "-based data structure)" else "off")
				^ "\ncaching: "
				^ (if !Settings.cachingEnabled then "on (using " ^ (case !Settings.cacheDatastructure of Settings.UBTREE => "UBtree" | Settings.BITMATRIX => "bitmatrix" | Settings.SIZEBOUNDED n => "size-bounded (" ^ (Int.toString n) ^ ") bitmatrix" | _ => Exn.unexpected "Prover.printSettings: unexpected datastructure for caching") ^ "-based data structure)" ^ (if !Settings.checkCacheOften then " with additional cache queries when propagating" else "") else "off")
				^ "\nboolean constraint propagation: "
				^ (if !Settings.bcpFullyEnabled then "eager" else if !Settings.bcpEnabled then "on" else "off")
				^ "\nearly conflict detection: "
				^ (if !Settings.ecdDisabled then "off" else "on")
				^ "\ndisjoint branching: "
				^ (if !Settings.semanticBranchingEnabled then if !Settings.semanticBranchingNgl then "with no-good lists" else "on" else "off")
				^ "\nlazy branching: "
				^ (if !Settings.lazyBranching then if !Settings.lazyWithBoxes then "on" else "propositional literals only" else if !Settings.lazyWithBoxes then "boxes only" else "off")
				^ "\nExpansion order: " ^ (Agenda.getOrder ())
				^ "\nDisjunct ordering: " ^ (String.implode (!Term.Catstore.listOrder))
				^ "\nDiamond ordering: " ^ (if !Settings.diadisjdep then "with disjunctions by dependencies" else (case !Settings.diaExpOrder of Settings.FIFO => "FIFO" | Settings.LIFO => "LIFO" | Settings.OLDWORLD => "old nodes first" | Settings.NEWWORLD => "new nodes first" | Settings.DEP => "minimal dependency first"))
				^ "\nDisjunction ordering: " ^ (if !Settings.diadisjdep then "with diamonds by dependencies" else (case !Settings.disjExpOrder of Settings.FIFO => "FIFO" | Settings.LIFO => "LIFO" | Settings.OLDWORLD => "old nodes first" | Settings.NEWWORLD => "new nodes first" | Settings.DEP => "minimal dependency first"))
				^ "\nformula: "
				^ (!formula)
				^ "\nfile name: "
				^ (!fileName)
				^ "\nindex: "
				^ (Int.toString (!index))
				^ "\n"
			)
		
		
		fun printSettingsList () =
			print (
				  (!fileName)
				^ ","
				^ (Int.toString (!index))
				^ ","
				^ (if (!negate) then "--negate" else "")
				^ ","
				^ (Agenda.getOrder ())
				^ ","
				^ (String.implode (!Term.Catstore.listOrder))
				^ ","
				^ (
					if !Settings.diadisjdep
					then "n/a"
					else
						case !Settings.diaExpOrder
							of Settings.OLDWORLD => "--dia-old"
							| Settings.NEWWORLD => "--dia-new"
							| Settings.LIFO => "--dia-lifo"
							| Settings.FIFO => "--dia-fifo"
							| Settings.DEP => "--dia-dep"
				  )
				^ ","
				^ (
					if !Settings.diadisjdep
					then "n/a"
					else
						case !Settings.disjExpOrder
							of Settings.OLDWORLD => "--disj-old"
							| Settings.NEWWORLD => "--disj-new"
							| Settings.LIFO => "--disj-lifo"
							| Settings.FIFO => "--disj-fifo"
							| Settings.DEP => "--disj-dep"
				  )
				^ ","
				^ (if !Settings.backjumpingDisabled then "--backjumping=off" else "--backjumping=on")
				^ ","
				^ (if !Settings.pbBlockingEnabled then (if !Settings.fullBlocking then "--blocking=eager" else "--blocking=on") else "--blocking=off")
				^ ","
				^ (if !Settings.cachingEnabled then if !Settings.checkCacheOften then "--caching=eager" else "--caching=on" else "--caching=off")
				^ ","
				^ (if !Settings.bcpFullyEnabled then "--bcp=eager" else if !Settings.bcpEnabled then "--bcp=on" else "--bcp=off")
				^ ","
				^ (if !Settings.ecdDisabled then "--ecd=off" else "--ecd=on")
				^ ","
				^ (if !Settings.semanticBranchingEnabled then if !Settings.semanticBranchingNgl then "--db=ngl" else "--db=on" else "--db=off")
				^ ","
				^ (if !Settings.lazyBranching then if !Settings.lazyWithBoxes then "--lazy=on" else "--lazy=prop" else if !Settings.lazyWithBoxes then "--lazy=box" else "--lazy=off")
				^ ","
				^ (
					case !Settings.pbbDatastructure
						of Settings.LISTS => "--blockingds=list"
						 | Settings.UBTREE => "--blockingds=tree"
						 | Settings.BITMATRIX => "--blockingds=matrix"
						 | _ => Exn.unexpected "Prover.printSettingsList: unexpected data structure for blocking."
				  )
				^ ","
				^ (
					case !Settings.cacheDatastructure
						of Settings.UBTREE => "--cachingds=tree"
						 | Settings.BITMATRIX => "--cachingds=matrix"
						 | Settings.SIZEBOUNDED n => "--cachingds=matrix:" ^ (Int.toString n)
						 | _ => Exn.unexpected "Prover.printSettingsList: unexpected data structure for caching."
				  )
				^ ","
				^ (
					case !Settings.timeout
						of NONE => ""
						 | SOME t => "--timeout=" ^ (Time.toString t)
				  )
				^ ","
			)
		
		fun printModel () =
			if !Settings.pbBlockingEnabled
			then
				let
					val xs = Nodestore.toOutputList ()
					
					fun getBlocked w =
						case BlockingMgr.listBlockedSuccessors w
							of nil => ""
							|  xs => ("blocked:      " ^ (Util.listToString (fn (r, w) => r ^ ":" ^ (Int.toString (Node.getId w))) (Listsort.sort (fn ((r1, w1), (r2, w2)) => Util.compare [fn () => String.compare (r1, r2), fn () => Int.compare (Node.getId w1, Node.getId w2)]) xs)) ^ "\n")
					
					fun af (w, s) = (print (s ^ (getBlocked w) ^ "------------------------------------------------------------\n"))
				in
					print (
						  "--MODEL-----------------------------------------------------\n"
						^ (case length xs of 1 => "1 node\n" | n => (Int.toString n) ^ " nodes\n")
						^ "------------------------------------------------------------\n"
					)
					; app af xs
				end
			else print (Nodestore.toString ())
		
		fun writeDot f =
			let
				val file = TextIO.openOut f
				
				val _ = TextIO.output (file, "digraph Model {\n")
			in
				if !Settings.pbBlockingEnabled
				then
					let
						fun getBlocked w =
							case BlockingMgr.listBlockedSuccessors w
								of nil => ""
								 | ws => concat (List.map (fn (r, w') => (Int.toString (Node.getId w)) ^ " -> " ^ (Int.toString (Node.getId (w'))) ^ "[label=\" " ^ r ^ " \", style=dashed];\n") ws)
						
						fun af w = TextIO.output (file, getBlocked w)
					in
						  TextIO.output (file, Nodestore.toDot ())
						; app af (Nodestore.listNodes ())
					end
				else TextIO.output (file, Nodestore.toDot ())
				; TextIO.output (file, "}\n")
				; TextIO.closeOut file
			end
			handle _ => print ("Unable to write model to file " ^ (Option.valOf (!dotFileName)) ^ "\n")
		
		val timer = Timer.startRealTimer ()
		
		fun run () =
			let
			in
				app readargs args;
				if !Settings.csv then printSettingsList () else printSettings ();
				if !Settings.csv then () else print "Parsing input\n";
				(
				case !format
					of ARG =>
						let
							val x = Parser.parseString (!formula)
						in
							case x
								of SOME p => parsetree := p
								|  NONE => Exn.error ("Unable to parse formula!")
						end
					|  NAT =>
						let
							val x = Parser.parse (!fileName)
						in
							case x
								of SOME p => parsetree := p
								|  NONE => Exn.error ("Unable to parse formula!")
						end
					|  LWB =>
						let
							val fs = LwbParser.parse (!fileName)
							
							val f = #2 (
								case (List.find (fn (n, _) => n = (!index)) fs)
								of SOME x => x
								|  NONE => Exn.error ("Index " ^ (Int.toString (!index)) ^ 
													" not found in lwb input file!"))
						in
							parsetree := f
						end
					|  KSATC =>
						let
							val r = KsatcParser.parse (!fileName)
						in
							case r
								of SOME pt => parsetree := pt
								|  NONE => Exn.error ("Unable to parse KSatC file")
						end
					|  ALC =>
						let
							val r = AlcParser.parse (!fileName)
						in
							case r
								of SOME pt => parsetree := pt
								|  NONE => Exn.error ("Unable to parse ALC file")
						end
					|  INTOHYLO =>
						let
							val r = IntohyloParser.parse (!fileName)
						in
							case r
								of SOME pt => parsetree := pt
								|  NONE => Exn.error ("Unable to parse InToHyLo file")
						end
					|  DIMACS =>
						let
							val r = DimacsParser.parse (!fileName)
						in
							case r
								of SOME pt => parsetree := pt
								|  NONE => Exn.error ("Unable to parse InToHyLo file")
						end
					|  DFG =>
						let
							val r = DfgParser.parse (!fileName)
						in
							case r
								of SOME pt => parsetree := pt
								|  NONE => Exn.error ("Unable to parse InToHyLo file")
						end
					|  DL98 =>
						let
							val fs = AdvancedTkbParser.parse (!fileName)
							handle _ => (
								  print "Advanced DL98 parser could not parse the given file. Trying other DL98 parser.\n"
								; map SOME (TkbParser.parse (!fileName))
							)
							
							val f = (
								case List.nth (fs, !index - 1)
									of SOME f => f
									 | NONE => (
										  print "Advanced DL98 parser did not find a TEST concept. Trying other DL98 parser.\n"
										; List.nth (TkbParser.parse (!fileName), !index - 1)
										)
								)
								handle Subscript => Exn.error ("Index " ^ (Int.toString (!index)) ^ 
										" not found in dl98 input file!")
						in
							parsetree := f
						end
					|  TANCS =>
						let
							val r = TancsParser.parse (!fileName)
						in
							case r
								of SOME pt => parsetree := pt
								|  NONE => Exn.error ("Unable to parse TANCS file")
						end
				);
				(if !negate then parsetree := (Parsetree.neg (!parsetree)) else ());
				print (if !Settings.csv then (Time.toString (Timer.checkRealTimer timer)) ^ "," else "Parse time: " ^ (Time.toString (Timer.checkRealTimer timer)) ^ " sec\n");
				if Solver.solveTree (!parsetree)
				then (
					  if !showModel
					  then (
						  if !Settings.detailedModelView then print (Translator.toString ()) else ()
						; printModel ()
					    )
					  else ()
					; (
						case !dotFileName
							of NONE => ()
							 | SOME f => writeDot f
					  )
				)
				else ()
			end
	in
		run ()
		handle Help => ()
		     | InvalidArgument => ()
			 | e => (if !Settings.csv then OS.Process.exit OS.Process.failure else raise e)
	end
