(*****************************************************************************
 *  Author:
 *    Daniel N. Gtzmann <dngoetzmann@googlemail.com>
 *
 *  Copyright:
 *     Daniel N. Gtzmann, 2009
 *
 *  Last modified:
 *    $Date: 2009-03-09 17:10:49 +0100 (Mon, 09 Mar 2009) $
 *    $Author: goetzmann $
 *    $Revision: 401 $
 *
 *  This file is part of Spartacus,
 *  the tableau prover for hybrid logic
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *****************************************************************************)


structure Node :> NODE =
	struct
		exception Satisfied
		
		datatype node =
			  W of {
				  id              : int
				, predId          : int option
				, creationDepcy   : Dependency.depcy
				, nominalDepcies  : Dependency.depcy list ref
				, pattern         : Termstore.store
				, propositions    : Propstore.pstore
				, nominals        : Propstore.pstore
				, diamonds        : Termstore.store
				, boxes           : Termstore.store
				, disjunctions    : Termstore.store
				, nglstore        : Termstore.store
				, lazystore       : Lazystore.store
				, lazyboxdisj     : Termstore.store
				, blockedDiamonds : (Term.index * int) list ref
				, branchPoints    : int list ref
				}
		
		
		exception Unsat of node * Dependency.depcy list
		exception CachedUnsat of node * Dependency.depcy list
		
		val nextId = ref 0
		
		
		fun reset () = nextId := 0
		
		
		fun foldin (w as W {id, pattern, propositions, nominals, diamonds, boxes, disjunctions, nglstore, lazystore, lazyboxdisj, ...}) (k, d) =
			let
				fun afLazyBranch (t, d') =
					Agenda.insert (
						  Agenda.BRANCH (id, t, Dependency.updDepth d' (Dependency.btDepth d))
						, Dependency.btDepth d
					)
				
				fun afLazyToLazystore (t, d') =
					if !Settings.lazyBranching andalso Lazystore.add lazystore (t, Dependency.updDepth d' (Dependency.btDepth d))
					then ()
					else afLazyBranch (t, d')
				
				fun foldin' k = (
					  (
						if !Settings.ecdDisabled
						then
							case Translator.getTerm k
								of Term.A (p, b) => (
									  (
										Propstore.add propositions p b d
										handle Propstore.Unsat (p, d, d') => raise Unsat (w, [d, d'])
									  )
									  ; if !Settings.lazyBranching
									    then app afLazyBranch (Lazystore.assert lazystore (p, b) (Dependency.btDepth d))
										else ()
									)
								|  Term.DMD _ => (
										if !Settings.lazyWithBoxes andalso Termstore.isEmpty diamonds
										then app afLazyToLazystore (Termstore.listItems lazyboxdisj)
										else ()
										;
										if Termstore.add diamonds (k, d)
										then Agenda.insert (Agenda.DIAMOND (id, k, d), Dependency.btDepth d)
										else ()
									)
								|  Term.BOX _ =>
									if Termstore.add boxes (k, d)
									then Agenda.insert (Agenda.BOX (id, k, d), Dependency.btDepth d)
									else ()
								|  Term.DISJ cs =>
									if Termstore.add disjunctions (k, d)
									then
										if
											!Settings.lazyWithBoxes
											andalso Termstore.isEmpty diamonds
											andalso (
												if (RelationMgr.someReflexive ())
												then
													List.exists
														(fn k' => case Translator.getTerm k
																of Term.BOX (r', _) => not (RelationMgr.isReflexive r')
																 | _ => Exn.unexpected "Node.foldin.foldin': not a box"
														)
														(Term.Catstore.getBoxes cs)
												else not (null (Term.Catstore.getBoxes cs))
											)
										then ignore (Termstore.add lazyboxdisj (k, d))
										else
											if !Settings.lazyBranching andalso Lazystore.add lazystore (k, d)
											then ()
											else Agenda.insert (Agenda.BRANCH (id, k, d), Dependency.btDepth d)
									else ()
								|  Term.ALL k' => if Universalstore.add (k', d) then Agenda.insert (Agenda.UNIV (k, d), Dependency.btDepth d) else ()
								|  Term.EX k' => if Existentialstore.add (k', d) then Agenda.insert (Agenda.EXIST (k', d), Dependency.btDepth d) else ()
								|  Term.CONJ cs => app foldin' (Term.Catstore.listItems cs)
								|  Term.AT (n, k') => Agenda.insert (Agenda.AT (n, k', d), Dependency.btDepth d)
								|  Term.EQ n => (
									  Propstore.add nominals n true d handle Propstore.Unsat (n, d1, d2) => raise Unsat (w, [d1, d2])
									; Agenda.insert (Agenda.EQ (id, n, d), Dependency.btDepth d)
									)
								|  Term.NEQ n => (
									Propstore.add nominals n false d
									handle Propstore.Unsat (n, d1, d2) => raise Unsat (w, [d1, d2])
									)
						else
							case Translator.getTerm k
								of Term.A (p, b) => (
									  (
										Propstore.add propositions p b d
										handle Propstore.Unsat (p, d, d') => raise Unsat (w, [d, d'])
									  )
									  ; if !Settings.lazyBranching
									    then app afLazyBranch (Lazystore.assert lazystore (p, b) (Dependency.btDepth d))
										else ()
									)
								|  Term.DMD _ => (
									case Termstore.peek boxes (Translator.getNegation k)
										of NONE => (
												if !Settings.lazyWithBoxes andalso Termstore.isEmpty diamonds
												then app afLazyToLazystore (Termstore.listItems lazyboxdisj)
												else ()
												;
												if Termstore.add diamonds (k, d)
												then Agenda.insert (Agenda.DIAMOND (id, k, d), Dependency.btDepth d)
												else ()
											)
										|  SOME d' => raise Unsat (w, [d, d'])
									)
								|  Term.BOX _ => (
									case Termstore.peek diamonds (Translator.getNegation k)
										of NONE =>
												if Termstore.add boxes (k, d)
												then Agenda.insert (Agenda.BOX (id, k, d), Dependency.btDepth d)
												else ()
										|  SOME d' => raise Unsat (w, [d, d'])
									)
								|  Term.DISJ cs => (
									if cs = Term.Catstore.empty
									then raise Unsat (w, [d])
									else
										case Termstore.peek pattern (Translator.getNegation k)
											of NONE =>
													if Termstore.add disjunctions (k, d)
													then
										if
											!Settings.lazyWithBoxes
											andalso Termstore.isEmpty diamonds
											andalso (
												if (RelationMgr.someReflexive ())
												then
													List.exists
														(fn k' => case Translator.getTerm k
																of Term.BOX (r', _) => not (RelationMgr.isReflexive r')
																 | _ => Exn.unexpected "Node.foldin.foldin': not a box"
														)
														(Term.Catstore.getBoxes cs)
												else not (null (Term.Catstore.getBoxes cs))
											)
														then ignore (Termstore.add lazyboxdisj (k, d))
														else
															if !Settings.lazyBranching andalso Lazystore.add lazystore (k, d)
															then ()
															else Agenda.insert (Agenda.BRANCH (id, k, d), Dependency.btDepth d)
													else ()
											|  SOME d' => raise Unsat (w, [d, d'])
									)
								|  Term.ALL k' => (
									case Existentialstore.peek k'
										of NONE => if Universalstore.add (k', d) then Agenda.insert (Agenda.UNIV (k', d), Dependency.btDepth d) else ()
										|  SOME d' => raise Unsat (w, [d, d'])
									)
								|  Term.EX k' => (
									case Universalstore.peek k'
										of NONE => if Existentialstore.add (k', d) then Agenda.insert (Agenda.EXIST (k', d), Dependency.btDepth d) else ()
										|  SOME d' => raise Unsat (w, [d, d'])
									)
								|  Term.CONJ cs => (
										case Termstore.peek disjunctions (Translator.getNegation k)
											of NONE => app foldin' (Term.Catstore.listItems cs)
											|  SOME d' => raise Unsat (w, [d, d'])
									)
								|  Term.AT (n, k') => Agenda.insert (Agenda.AT (n, k', d), Dependency.btDepth d)
								|  Term.EQ n => (
									  Propstore.add nominals n true d handle Propstore.Unsat (n, d1, d2) => raise Unsat (w, [d1, d2])
									; Agenda.insert (Agenda.EQ (id, n, d), Dependency.btDepth d)
									)
								|  Term.NEQ n =>  (
									Propstore.add nominals n false d
									handle Propstore.Unsat (n, d1, d2) => raise Unsat (w, [d1, d2])
									)
					  )
					; (
						if !Settings.semanticBranchingNgl
						then (case Termstore.peek nglstore k of SOME d' => (Debug.output (fn () => "Semantic branching: Conflict with NGL item " ^ (Int.toString k) ^ "\n"); raise Unsat (w, [d, d'])) | _ => ())
						else ()
					  )
				)
			in
				foldin' k
			end
		
		
		fun newNode curDepth predId (k, d) =
			let
				val d' = Dependency.updDepth d curDepth
				
				val pstore = Propstore.empty ()
				
				val w =	W {
						  id              = !nextId
						, predId          = predId
						, creationDepcy   = d'
						, nominalDepcies  = ref nil
						, pattern         = let val pattern = Termstore.empty () in Termstore.add pattern (k, d'); pattern end
						, propositions    = pstore
						, nominals        = Propstore.empty ()
						, diamonds        = Termstore.empty ()
						, boxes           = Termstore.empty ()
						, disjunctions    = Termstore.empty ()
						, nglstore        = Termstore.empty ()
						, lazystore       = Lazystore.empty pstore
						, lazyboxdisj     = Termstore.empty ()
						, blockedDiamonds = ref nil
						, branchPoints    = ref nil
						}
				in
					w before (Ref.incr nextId; foldin w (k, d'))
			end
		
		
		fun getId (W {id, ...}) = id
		
		
		fun peekPredId (W {predId, ...}) = predId
		
		
		fun propagate (w as W {id, creationDepcy, nominalDepcies, pattern, ...}) curDepth xs =
			let
				fun propagate' (k, d) =
					let
						val d' =
							Dependency.updDepth (Dependency.union (d, hd (!nominalDepcies)
								handle List.Empty => creationDepcy)) curDepth
					in
						  Termstore.add pattern (k, d')
						; foldin w (k, d')
					end
			in
				  Debug.output (fn () => "Propagating " ^ (Util.listToString (fn (k, _) => Int.toString k) xs) ^ " to node " ^ (Int.toString id) ^ "\n")
				; (
					if !Settings.cachingEnabled andalso !Settings.checkCacheOften andalso Cache.isCachedUnsat ((map #1 xs)@(map #1 (Termstore.listItems pattern)))
					then (
						  Debug.output
							(fn () => "Pattern " ^ (Util.listToString (fn (k, _) => Int.toString k) xs)
								^ ", " ^ (Util.listToString (fn (k, _) => Int.toString k) (Termstore.listItems pattern))
								^ " is cached as unsatisfiable\n")
						; raise CachedUnsat (w, (map #2 xs)@(map #2 (Termstore.listItems pattern)))
					)
					else ()
				  )
				; app propagate' xs
			end
		
		
		fun addNominalDepcy (w as W {id, nominals, creationDepcy, nominalDepcies, ...}) curDepth d =
			let
				val d' = (Dependency.updDepth (Dependency.union
					(d, hd (!nominalDepcies) handle List.Empty => creationDepcy)) curDepth)
			in
				  Ref.push nominalDepcies d'
			end
		
		
		fun blockDiamond (w as W {blockedDiamonds, ...}) d k =
			Ref.push blockedDiamonds (k, d)
		
		
		fun unblockDiamonds (W {id, diamonds, blockedDiamonds, ...}) pf =
			let
				fun unblockDiamond (k, _) =
					let
						val d =
							case Termstore.peek diamonds k
								of NONE => Exn.unexpected "Termstore.unblock: diamond to be unblocked not stored."
								 | SOME d => d
					in
						Agenda.insert (Agenda.DIAMOND (id, k, d), Dependency.btDepth d)
					end
				
				val (remaining, unblocked) =
					List.partition pf (!blockedDiamonds)
			in
				  Debug.output (fn () => (Int.toString (length unblocked)) ^ " out of " ^ (Int.toString ((length unblocked) + (length remaining))) ^ " diamonds are no longer blocked\n")
				; app unblockDiamond unblocked
				; blockedDiamonds := remaining
			end
		
		
		fun listBlockedDiamonds (W {blockedDiamonds, ...}) = map #1 (!blockedDiamonds)
		
		
		fun listBranchPoints (W {branchPoints, ...}) = !branchPoints
		
		
		fun propagateDisjunct (w as W {id, nglstore, propositions, disjunctions, diamonds, boxes, nominals, branchPoints, ...}) (k, d, ngs, more) =
			let
				fun checkNgl k = (
					case Translator.getTerm k
						of Term.A (p, b) => (
							case Propstore.peek propositions p
								of SOME (b', d') => if b = b' then raise Unsat (w, [d, d']) else ()
								 | _ => ()
							)
						|  Term.EQ n => (
							case Propstore.peek nominals n
								of SOME (b, d') => if b then raise Unsat (w, [d, d']) else ()
								 | _ => ()
							)
						|  Term.NEQ n => (
							case Propstore.peek nominals n
								of SOME (b, d') => if b then () else raise Unsat (w, [d, d'])
								 | _ => ()
							)
						|  Term.DISJ _ => (
							case Termstore.peek disjunctions k
								of SOME d' => raise Unsat (w, [d, d'])
								 | _ => ()
							)
						|  Term.DMD _ => (
							case Termstore.peek diamonds k
								of SOME d' => raise Unsat (w, [d, d'])
								 | _ => ()
							)
						|  Term.BOX _ => (
							case Termstore.peek boxes k
								of SOME d' => raise Unsat (w, [d, d'])
								 | _ => ()
							)
						|  _ => ()
				) handle (e as Unsat _) => (Debug.output (fn () => "Semantic Branching: Conflict with new NGL item\n"); raise e)
			in
				  if more then Ref.push branchPoints (Dependency.btDepth d) else ()
				; app (fn k => ignore (Termstore.add nglstore (k, d))) ngs
				; foldin w (k, d)
				; (
					if !Settings.semanticBranchingEnabled
					then if !Settings.semanticBranchingNgl
					then app checkNgl ngs
					else app (fn k => foldin w (Translator.getNegation k, d)) ngs
					else ()
				  )
			end
		
		
		fun getPattern (W {pattern, ...}) = Termstore.listItems pattern
		
		fun listBoxes (W {boxes, ...}) = Termstore.listItems boxes
		
		
		fun listPropositions (W {propositions, lazystore, ...}) =
			if !Settings.lazyBranching
			then 
				Listsort.sort
					(fn ((x, _), (y, _)) => String.compare (x, y))
					((map (fn x => (#1 x, #2 x)) (Propstore.listItems propositions)) @ (Lazystore.listPropositions lazystore))
			else map (fn x => (#1 x, #2 x)) (Propstore.listItems propositions)
		
		
		fun listNominals (W {nominals, ...}) = map (fn x => (#1 x, #2 x)) (Propstore.listItems nominals)
		
		
		fun bcp (W {pattern, propositions, nominals, diamonds, boxes, disjunctions, nglstore, ...}) cs =
			let
				val dsr = ref nil
				
				fun fDiamonds k =
					case Termstore.peek diamonds k
						of SOME _ => raise Satisfied
						|  NONE => (
							case Termstore.peek boxes (Translator.getNegation k)
								of NONE => true
								|  SOME d => (Ref.push dsr d; false)
							)
				
				fun fBoxes k =
					case Termstore.peek boxes k
						of SOME _ => raise Satisfied
						|  NONE => (
							case Termstore.peek diamonds (Translator.getNegation k)
								of NONE => true
								|  SOME d => (Ref.push dsr d; false)
							)
				
				fun fConjunctions k =
					case Termstore.peek pattern k
						of SOME _ => raise Satisfied
						|  NONE => (
							case Termstore.peek disjunctions (Translator.getNegation k)
								of NONE => true
								|  SOME d => (Ref.push dsr d; false)
							)
				
				fun fDisjunctions k =
					case Termstore.peek disjunctions k
						of SOME _ => raise Satisfied
						|  NONE => (
							case Termstore.peek pattern (Translator.getNegation k)
								of NONE => true
								|  SOME d => (Ref.push dsr d; false)
							)
				
				fun fUniversals k =
					case Universalstore.peek k
						of SOME _ => raise Satisfied
						|  NONE => (
							case Existentialstore.peek (Translator.getNegation k)
								of NONE => true
								|  SOME d => (Ref.push dsr d; false)
							)
				
				fun fExistentials k =
					case Existentialstore.peek k
						of SOME _ => raise Satisfied
						|  NONE => (
							case Universalstore.peek (Translator.getNegation k)
								of NONE => true
								|  SOME d => (Ref.push dsr d; false)
							)
				
				fun ff k =
					let
						fun ffa k =
							case Translator.getTerm k
								of Term.A (p, b) => (
									case Propstore.peek propositions p
										of NONE => true
										|  SOME (b', d) =>
											if b = b'
											then raise Satisfied
											else (Ref.push dsr d; false)
									)
								|  Term.DMD _ => fDiamonds k
								|  Term.BOX _ => fBoxes k
								|  Term.CONJ _ => fConjunctions k
								|  Term.DISJ _ => fDisjunctions k
								|  Term.ALL k' => fUniversals k'
								|  Term.EX k' => fExistentials k'
								|  Term.EQ n => (
										case Propstore.peek nominals n
											of NONE => true
											 | SOME (true, d) => raise Satisfied
											 | SOME (false, d) => (Ref.push dsr d; false)
									)
								| Term.NEQ n => (
										case Propstore.peek nominals n
											of NONE => true
											 | SOME (true, d) => (Ref.push dsr d; false)
											 | SOME (false, d) => raise Satisfied
									)
								|  _ => true
						
						fun ffb k =
							case Termstore.peek nglstore k
								of NONE => true
								|  SOME d => (Ref.push dsr d; false)
					in
						ffa k andalso (not (!Settings.semanticBranchingNgl) orelse ffb k)
					end
			in
				(Term.Catstore.filter ff cs, !dsr)
			end
		
		
		fun mergeInto
				(w1 as W {nominalDepcies = r_nd, pattern = r_pat, propositions = r_props, nominals = r_noms, diamonds = r_dias, boxes = r_boxes, disjunctions = r_disj, lazystore = r_lazystore, lazyboxdisj = r_lazyboxdisj, id, ...})
				curDepth
				(w2 as W {nominalDepcies = o_nd, pattern = o_pat, propositions = o_props, nominals = o_noms, diamonds = o_dias, boxes = o_boxes, disjunctions = o_disj, lazystore = o_lazystore, lazyboxdisj = o_lazyboxdisj, ...})
				d =
					let
						val d' = Dependency.updDepth (Dependency.union (hd (!r_nd), hd (!o_nd))) curDepth
						
						val mustHandleLazyBoxes =
							!Settings.lazyWithBoxes
							andalso Termstore.isEmpty r_dias
							andalso not (Termstore.isEmpty o_dias)
						
						fun afLazyBranch (t, d') =
							Agenda.insert (
								  Agenda.BRANCH (id, t, Dependency.updDepth d' curDepth)
								, curDepth
							)
						
						fun afLazy (k, d) =
							let
								val d = Dependency.union (d, d')
							in
								case Translator.getTerm k
									of Term.DISJ cs =>
										if !Settings.lazyWithBoxes andalso Termstore.isEmpty r_dias andalso (not (null (Term.Catstore.getBoxes cs)))
										then ignore (Termstore.add r_lazyboxdisj (k, d))
										else
											if !Settings.lazyBranching andalso Lazystore.add r_lazystore (k, d)
											then ()
											else Agenda.insert (Agenda.BRANCH (id, k, d), Dependency.btDepth d)
									 | _ => Exn.unexpected "Node.mergeInto.afLazy"
							end
					in
						  Ref.push r_nd d'
						; app (fn (k, d) => ignore (Termstore.add r_pat (k, Dependency.union (d, d')))) (Termstore.listItems o_pat)
						; app
							(fn (p, b, d) => (
								  ignore (Propstore.add r_props p b (Dependency.union (d, d')))
								; if !Settings.lazyBranching
								  then app afLazyBranch (Lazystore.assert r_lazystore (p, b) (curDepth))
								  else ()
								)
							)
							(Propstore.listItems o_props)
						  handle Propstore.Unsat (p, d, d') => raise Unsat (w1, [d, d'])
						; app (fn (n, b, d) => ignore (Propstore.add r_noms n b (Dependency.union (d, d')))) (Propstore.listItems o_noms)
						  handle Propstore.Unsat (n, d1, d2) => raise Unsat (w1, [d1, d2])
						; app (fn (k, d) => ignore (Termstore.add r_dias (k, Dependency.union (d, d')))) (Termstore.listItems o_dias)
						; app (fn (k, d) => ignore (Termstore.add r_boxes (k, Dependency.union (d, d')))) (Termstore.listItems o_boxes)
						; app (fn (k, d) => ignore (Termstore.add r_disj (k, Dependency.union (d, d')))) (Termstore.listItems o_disj)
						; app afLazy (Lazystore.listItems o_lazystore)
						; app afLazy (Termstore.listItems o_lazyboxdisj)
						; if mustHandleLazyBoxes
						  then app afLazy (Termstore.listItems r_lazyboxdisj)
						  else ()
						; app (fn (k, d) => Agenda.insertX (Agenda.BOX (id, k, Dependency.union (d, d')), curDepth)) (Termstore.listItems r_boxes)
					end
		
		
		fun backtrack (w as W {id, nominalDepcies, pattern, propositions, nominals, diamonds, boxes, disjunctions, nglstore, lazystore, lazyboxdisj, blockedDiamonds, branchPoints, ...}) d =
			let
				fun bdmf nil = nil
				  | bdmf (xs as (_, d')::xr) = if d' <= d then xs else bdmf xr
				
				fun bpmf nil = nil
				  | bpmf (xs as (x::xr)) = if x <= d then xs else bpmf xr
			in
				  Termstore.backtrack pattern d
				; Ref.modify (List.filter (fn d' => Dependency.btDepth d' <= d)) nominalDepcies
				; Propstore.backtrack propositions d
				; Propstore.backtrack nominals d
				; Termstore.backtrack diamonds d
				; Termstore.backtrack boxes d
				; Termstore.backtrack disjunctions d
				; Termstore.backtrack nglstore d
				; Lazystore.backtrack lazystore d
				; Termstore.backtrack lazyboxdisj d
				; Ref.modify bdmf blockedDiamonds
				; Ref.modify bpmf branchPoints
			end
		
		
		fun toString (W {id, pattern, propositions, nominals, diamonds, boxes, disjunctions, nglstore, ...}) =
			  "\n-------NODE-------"
			^ "\n id: " ^ (Int.toString id)
			^ "\n nominals:     " ^ (Propstore.toString nominals)
			^ "\n pattern:      " ^ (Termstore.toString pattern)
			^ "\n propositions: " ^ (Propstore.toString propositions)
			^ "\n boxes:        " ^ (Termstore.toString boxes)
			^ "\n diamonds:     " ^ (Termstore.toString diamonds)
			^ "\n disjunctions: " ^ (Termstore.toString disjunctions)
			^ "\n ngl:          " ^ (Termstore.toString nglstore)
			^ "\n-------------------\n\n"
	end
